@@ -64,16 +64,15 @@
 ;;;    so it works in a sense -- there is a memory leak here
 ;;; 
 ;;;
-;;; (c-define (list '(void* calloc (size_t size_t))
-;;;		    '(void* malloc (size_t))
-;;;		    '(void free (void*))
-;;;		    '(void* realloc(void* size_t))
-;;;		    '(void time (time_t*)) ; ignore returned value
+;;; (c-define (list '(time_t* calloc (size_t size_t))
+;;;		    '(void free (time_t*))
+;;;		    '(void time (time_t*))
 ;;;		    (list (symbol "struct tm*") 'localtime '(time_t*))
 ;;;                 (list 'size_t 'strftime (list 'char* 'size_t 'char* (symbol "struct tm*"))))
 ;;;          "" "time.h")
-;;;   > (let ((p (calloc 1 8)) (str (make-string 32))) (time p) (strftime str 32 "%a %d-%b-%Y %H:%M %Z" (localtime p)) (free p) str)
-;;;   "Sat 11-Aug-2012 08:55 PDT\x00      "
+;;;   > (let ((p (calloc 1 8)) (str (make-string 32))) 
+;;;       (time p) (let ((bytes (strftime str 32 "%a %d-%b-%Y %H:%M %Z" (localtime p)))) (free p) (substring str 0 bytes)))
+;;;   "Sat 11-Aug-2012 08:55 PDT
 ;;;
 ;;;
 ;;; (c-define '((int closedir (DIR*))
@@ -297,7 +296,7 @@
 	(let* ((name (symbol->string type))
 	       (pos (char-position #\- name)))
 	  (if (not pos)
-	      type
+	      name ;type
 	      (begin
 		(string-set! name pos #\space)
 		(set! pos (char-position #\- name (+ pos 1)))
@@ -312,8 +311,12 @@
 		      (len (length name)))
 		 (do ((i 0 (+ i 1)))
 		     ((= i len))
-		   (if (char=? (name i) #\*) (set! (name i) #\_))
-		   (if (char=? (name i) #\-) (set! (name i) #\_)))
+		   (if (char=? (name i) #\*) 
+		       (set! (name i) #\_)
+		       (if (char=? (name i) #\-) 
+			   (set! (name i) #\_)
+			   (if (char=? (name i) #\space) 
+			       (set! (name i) #\_))))) ; "struct tm*" for example
 		 (set! name (symbol name "_symbol"))
 		 (set! type-symbols (cons (cons type name) type-symbols))
 		 name))))
@@ -325,7 +328,6 @@
 	  (let ((num-args (length arg-types))
 		(base-name (string-append (if (> (length prefix) 0) prefix "s7_") "_" func-name)) ; not "g" -- collides with glib
 		(scheme-name (string-append prefix (if (> (length prefix) 0) ":" "") func-name)))
-	  
 	    (if (and (= num-args 1) 
 		     (eq? (car arg-types) 'void))
 		(set! num-args 0))
@@ -344,11 +346,9 @@
 	      (do ((i 0 (+ i 1))
 		   (type arg-types (cdr type)))
 		  ((= i num-args))
-		
 		(let* ((nominal-type ((if (pair? (car type)) caar car) type))  ; double in the example
 		       (true-type    ((if (pair? (car type)) cadar car) type))
-		       (s7-type      (C-type->s7-type true-type)))                    ; real
-
+		       (s7-type      (C-type->s7-type true-type)))             ; real
 		  (format pp "  arg = s7_car(p);~%")
 		  (if (eq? true-type 's7_pointer)
 		      (format pp "    ~A_~D = arg;~%" base-name i)
@@ -498,43 +498,45 @@
 	(format p "{~%")
 	(format p "  s7_pointer cur_env;~%")
 
-	(format p "  s7_pointer ")
-	(let ((pls (hash-table-entries signatures))
-	      (loc 1))
-	  (for-each
-	   (lambda (s)
-	     (format p "~A~A~A" (cdr s) (if (< loc pls) (values "," " ") (values ";" #\newline)))
-	     (set! loc (+ loc 1)))
-	   signatures))
-	
-	(let ((syms ())
-	      (names ()))
-	  (for-each
-	   (lambda (q)
-	     (when (positive? (cdr q))
-	       (set! syms (cons (car q) syms))
-	       (set! names (cons (signature->pl (car q)) names))))
-	   sig-symbols)
-	  (when (pair? syms)
-	    (format p "  {~%    s7_pointer ~{~C~^, ~};~%" names)
+	(when (> (hash-table-entries signatures) 0)   ; maybe just constants (no functions)
+	  (format p "  s7_pointer ")
+	  (let ((pls (hash-table-entries signatures))
+		(loc 1))
 	    (for-each
-	     (lambda (name sym)
-	       (format p (if (eq? sym 't)
-			     "    t = s7_t(sc);~%"
-			     (values "    ~C = s7_make_symbol(sc, ~S);~%" name (symbol->string sym)))))
-	     names syms)))
-	(format p "~%")
-	(for-each
-	 (lambda (sig)
-	   (let ((siglen (length (cdar sig)))
-		 (cyclic (char=? ((cdr sig) 1) #\c)))
-	     (format p (if cyclic 
-			   (values "    ~A = s7_make_circular_signature(sc, ~D, ~D" (cdr sig) (- siglen 1) siglen)
-			   (values "    ~A = s7_make_signature(sc, ~D" (cdr sig) siglen)))
-	     (format p "~{~^, ~C~}" (substring (cdr sig) (if cyclic 4 3)))
-	     (format p ");~%")))
-	 signatures)
-	(format p "  }~%~%")
+	     (lambda (s)
+	       (format p "~A~A~A" (cdr s) (if (< loc pls) (values "," " ") (values ";" #\newline)))
+	       (set! loc (+ loc 1)))
+	     signatures))
+	  
+	  (let ((syms ())
+		(names ()))
+	    (for-each
+	     (lambda (q)
+	       (when (positive? (cdr q))
+		 (set! syms (cons (car q) syms))
+		 (set! names (cons (signature->pl (car q)) names))))
+	     sig-symbols)
+	    (when (pair? syms)
+	      (format p "  {~%    s7_pointer ~{~C~^, ~};~%" names)
+	      (for-each
+	       (lambda (name sym)
+		 (format p (if (eq? sym 't)
+			       "    t = s7_t(sc);~%"
+			       (values "    ~C = s7_make_symbol(sc, ~S);~%" name (symbol->string sym)))))
+	       names syms)))
+	  (format p "~%")
+	  (for-each
+	   (lambda (sig)
+	     (let ((siglen (length (cdar sig)))
+		   (cyclic (char=? ((cdr sig) 1) #\c)))
+	       (format p (if cyclic 
+			     (values "    ~A = s7_make_circular_signature(sc, ~D, ~D" (cdr sig) (- siglen 1) siglen)
+			     (values "    ~A = s7_make_signature(sc, ~D" (cdr sig) siglen)))
+	       (format p "~{~^, ~C~}" (substring (cdr sig) (if cyclic 4 3)))
+	       (format p ");~%")))
+	   signatures)
+	  (format p "  }~%~%"))
+
 	(format p "  cur_env = s7_curlet(sc);~%") ; changed from s7_outlet(s7_curlet) 20-Aug-17
 	
 	;; send out any special initialization code
@@ -575,29 +577,28 @@
 	       constants)))
 	
 	;; C macros -- need #ifdef name #endif wrapper
-	(if (pair? macros)
-	    (begin
-	      (format p "~%")
-	      (for-each
-	       (lambda (c)
-		 (let* ((type (c 0))
-			(c-name (c 1))
-			(scheme-name (string-append prefix (if (> (length prefix) 0) ":" "") c-name))
-			(trans (C->s7 type)))
-		   (format p "#ifdef ~A~%" c-name)
-		   (if (eq? trans 's7_make_c_pointer_with_type)
-		       (format p "  s7_define(sc, cur_env, s7_make_symbol(sc, ~S), s7_make_c_pointer_with_type(sc, (~A)~A, s7_make_symbol(sc, \"~S\"), s7_f(sc)));~%" 
-			       scheme-name
-			       (C->s7-cast type)
-			       c-name
-			       type)
-		       (format p "  s7_define(sc, cur_env, s7_make_symbol(sc, ~S), ~A(sc, (~A)~A));~%" 
-			       scheme-name
-			       trans
-			       (C->s7-cast type)
-			       c-name))
-		   (format p "#endif~%")))
-	       macros)))
+	(when (pair? macros)
+	  (format p "~%")
+	  (for-each
+	   (lambda (c)
+	     (let* ((type (c 0))
+		    (c-name (c 1))
+		    (scheme-name (string-append prefix (if (> (length prefix) 0) ":" "") c-name))
+		    (trans (C->s7 type)))
+	       (format p "#ifdef ~A~%" c-name)
+	       (if (eq? trans 's7_make_c_pointer_with_type)
+		   (format p "  s7_define(sc, cur_env, s7_make_symbol(sc, ~S), s7_make_c_pointer_with_type(sc, (~A)~A, s7_make_symbol(sc, \"~S\"), s7_f(sc)));~%" 
+			   scheme-name
+			   (C->s7-cast type)
+			   c-name
+			   type)
+		   (format p "  s7_define(sc, cur_env, s7_make_symbol(sc, ~S), ~A(sc, (~A)~A));~%" 
+			   scheme-name
+			   trans
+			   (C->s7-cast type)
+			   c-name))
+	       (format p "#endif~%")))
+	   macros))
 	
 	;; functions
 	(for-each
@@ -650,7 +651,7 @@
 		    (not (pair? *load-path*)))
 	  (set! *cload-cflags* (append *cload-cflags* (format #f " -I~A" (car *load-path*)))))
 	
-	;; now we have the module .c file -- make it into a shared object, load it, delete the temp files
+	;; now we have the module .c file -- make it into a shared object
 	
 	(cond ((provided? 'osx)
 	       ;; I assume the caller is also compiled with these flags?
@@ -677,7 +678,13 @@
 	       (system (format #f "cc ~A -G -o ~A ~A ~A" 
 			       o-file-name so-file-name *cload-ldflags* ldflags)))
 	      
-	      (else
+	      ((or (provided? 'mingw) (provided? 'msys2)) ; from chai xiaoxiang
+	       ;; you'll need dlfcn which can be installed with pacman, and remember to build s7 with -DWITH_C_LOADER=1
+	       ;; in msys2:  gcc s7.c -o s7 -DWITH_MAIN -DWITH_C_LOADER=1 -I. -O2 -g -ldl -lm -Wl,-export-all-symbols,--out-implib,s7.lib
+	       (system (format #f "gcc ~A s7.lib -shared -o ~A -I. ~A ~A"
+			       c-file-name so-file-name cflags ldflags)))
+	       
+	      (else ; linux netbsd
 	       (system (format #f "~A -fPIC -c ~A -o ~A ~A ~A" 
 			       *cload-c-compiler* c-file-name o-file-name *cload-cflags* cflags))
 	       (system (format #f "~A ~A -shared -o ~A ~A ~A" 
